﻿using MathNet.Numerics.Random;
using Microsoft.ML;
using Microsoft.ML.Data;
using PLplot;
using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.ML.Data;

class Program
{
    static void Main(string[] args)
    {
        MLContext ctx = new MLContext();

        IDataView data = ctx.Data.LoadFromTextFile<ClickData>("click.csv", separatorChar: ',', hasHeader: true);

        // 표준화
        var xyList = ctx.Data.CreateEnumerable<ClickData>(data, false).NormalizeZscore();

        // 매개변수 초기화
        double theta0 = SystemRandomSource.Default.NextDouble();
        double theta1 = SystemRandomSource.Default.NextDouble();

        // 예측 함수
        Func<double, double> f = (x) => theta0 + theta1 * x;

        // 목적 함수
        Func<double, double, double> errorFunc = (x, y) => Math.Pow((y - f(x)), 2);
        Func<IEnumerable<ClickData>, double> E = (list) => 0.5 * list.ForEach((e) => errorFunc(e.X, e.Y)).Sum();

        // 학습률
        double ETA = 1e-03;

        // 오차의 차분
        double diff = 1.0;

        // 갱신 횟수
        int count = 0;

        // 오차의 차분이 0.01 이하가 될 떄까지 매개변수 갱신을 반복
        double error = E(xyList);

        while (diff > 1e-02)
        {
            // 갱신 결과를 임시 변수에 저장
            double tmp_theta0 = theta0 - ETA * xyList.ForEach((e) => f(e.X) - e.Y).Sum();
            double tmp_theta1 = theta1 - ETA * xyList.ForEach((e) => (f(e.X) - e.Y) * e.X).Sum();

            // 매개변수 갱신
            theta0 = tmp_theta0;
            theta1 = tmp_theta1;

            // 이전 회의 오차와의 차분을 계산
            double currentError = E(xyList);
            diff = error - currentError;
            error = currentError;

            // 로그 출력
            count++;
            Console.WriteLine($"{count,4:#} 회째: theta0 = {theta0,8:#.0000}, theta1 = {theta1,8:#.0000}, 차분 = {diff,8:#.0000}");
        }

        // 그래프 출력
        double[] xData = xyList.Select((elem) => elem.X).ToArray();
        double[] yData = xyList.Select((elem) => elem.Y).ToArray();
        DrawPlotChart(xData, yData, f);
    }

    private static void DrawPlotChart(double[] xData, double[] yData, Func<double, double> drawFunc)
    {
        string chartFileName = "click.svg";

        double xMin = xData.Min();
        double yMin = yData.Min();

        double xMax = xData.Max();
        double yMax = yData.Max();

        using (var pl = new PLStream())
        {
            pl.sdev("svg");
            pl.sfnam(chartFileName);
            pl.spal0("cmap0_alternate.pal");
            pl.init();

            pl.env(xMin, xMax, yMin, yMax, AxesScale.Independent, AxisBox.BoxTicksLabelsAxes);
            pl.lab("X", "Y", "Click");
            char code = Symbol.Bullet;
            pl.col0(2);

            pl.poin(xData, yData, code);

            {
                double[] ptX = Utils.RangeInclusive(xMin, xMax, 0.01).ToArray();
                double[] ptY = null;

                ptY = new double[ptX.Length];
                for (int i = 0; i < ptX.Length; i++)
                {
                    ptY[i] = drawFunc(ptX[i]);
                }

                pl.line(ptX, ptY);
            }

            pl.eop();
            pl.gver(out var verText);
        }
    }
}
